/*
 * This file is derived from the following files under the mit/lib/Xt 
 * directory of MIT X11R5 distribution.
 *
 * $XConsortium: TMparse.c,v 1.123 91/07/28 11:52:14 swick Exp $ 
 * $XConsortium: TMstate.c,v 1.158 91/08/21 11:04:40 converse Exp $ 
 * $XConsortium: TMkey.c,v 1.12 91/07/05 15:04:12 rws Exp $ 
 * $XConsortium: Alloc.c,v 1.46 91/07/30 11:04:41 rws Exp $
 */

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/*
****************************************************************************

              Copyright 1991, by FUJITSU LIMITED
              Copyright 1991, by Sun Microsystems, Inc.	      

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of FUJITSU LIMITED and Sun
Microsystems, Inc. not be used in advertising or publicity pertaining
to distribution of the software without specific, written prior
permission.  FUJITSU LIMITED and Sun Microsystems, Inc. makes no
representations about the suitability of this software for any
purpose.  It is provided "as is" without express or implied warranty.

FUJITSU LIMITED AND SUN MICROSYSTEMS, INC. DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL FUJITSU LIMITED AND SUN
MICROSYSTEMS, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

****************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#ifdef SVR4
#include <sys/systeminfo.h>
#endif /* SVR4 */
#include <sys/types.h>
#include <sys/stat.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#if XlibSpecificationRelease != 5
#include <X11/XlibR5.h>
#endif /* XlibSpecificationRelease */
#include "XSunExt.h"
#include <X11/keysym.h>
#include <X11/Xutil.h>

#include "Htt.h"
#include "XSunIMProt.h"
#include "XSunIMPriv.h"
#include "XSunIMMMan.h"
#include "XSunIMMthd.h"
#include "XSunIMPub.h"
/*#include "XSunIMCore.h" included in Htt.h*/
#include "XSunIMTrans.h"
#include "XSunIMTransP.h"

#define DISPLAY_SUPPORT

#ifndef	XtMalloc
#define	XtMalloc(size)		_HttMalloc(size)
#endif
#ifndef	XtNew
#define	XtNew(type)		((type *)XtMalloc(sizeof(type)))
#endif
#ifndef XtRealloc
#define XtRealloc(ptr,size)	_HttRealloc(ptr,size)
#endif
#ifndef XtFree
#define XtFree(ptr)		_HttFree(ptr)
#endif
#ifndef MIN
#define	MIN(a,b)		((a) < (b) ? (a) : (b) )
#endif

#ifdef SVR4
#ifndef bzero
#define	bzero(ptr,size)		memset((ptr),'\0',size)
#endif
#ifndef bcopy
#define	bcopy(in,out,size)	memcpy(out,in,size)
#endif
#endif /* SVR4 */

#ifndef XtBCopy
#define	XtBCopy(in,out,size)	bcopy(in,out,size)
#endif

#define XtWarningMsg(arg1,arg2,arg3,arg4,arg5,arg6)

#define TMGetActionQuark(quarkTbl,idx)	((quarkTbl)[idx])

static String XtNtranslationParseError = (String)"translationParseError";

typedef int		EventType;

typedef String (*ParseProc)(); /* str, closure, event ,error */
    /* String str; */
    /* Opaque closure; */
    /* EventPtr event; */
    /* Boolean* error */

typedef void (*ModifierProc)(); 
typedef TMShortCard	Value;
typedef long		Opaque;

typedef struct _ModifierRec {
    char*      name;
    XrmQuark   signature;
    ModifierProc modifierParseProc;
    Value      value;
} ModifierRec, *ModifierKeys;

typedef struct _EventKey {
    char    	*event;
    XrmQuark	signature;
    EventType	eventType;
    ParseProc	parseDetail;
    Opaque	closure;
}EventKey, *EventKeys;

typedef struct _NameValueRec {
    char	*name;
    XrmQuark	 signature;
    Value	 value;
} NameValueRec, *NameValueTable;

Private void ParseModImmed();
Private void ParseModSym();
Private String PanicModeRecovery();
Private String CheckForPoundSign();
Private KeySym StringToKeySym();
Private void CallMatchActions();

#ifndef DISPLAY_SUPPORT
extern Display	*display;
#endif

static ModifierRec modifiers[] = {
    {"None",    0,      ParseModImmed,None},
    {"Shift",	0,	ParseModImmed,ShiftMask},
    {"Lock",	0,	ParseModImmed,LockMask},
    {"Ctrl",	0,	ParseModImmed,ControlMask},
    {"Mod1",	0,	ParseModImmed,Mod1Mask},
    {"Mod2",	0,	ParseModImmed,Mod2Mask},
    {"Mod3",	0,	ParseModImmed,Mod3Mask},
    {"Mod4",	0,	ParseModImmed,Mod4Mask},
    {"Mod5",	0,	ParseModImmed,Mod5Mask},
    {"Meta",	0,	ParseModSym,  XK_Meta_L},
    {"m",       0,      ParseModSym,  XK_Meta_L},
    {"h",       0,      ParseModSym,  XK_Hyper_L},
    {"su",      0,      ParseModSym,  XK_Super_L},
    {"a",       0,      ParseModSym,  XK_Alt_L},
    {"Hyper",   0,      ParseModSym,  XK_Hyper_L},
    {"Super",   0,      ParseModSym,  XK_Super_L},
    {"Alt",     0,      ParseModSym,  XK_Alt_L},
    {"Button1",	0,	ParseModImmed,Button1Mask},
    {"Button2",	0,	ParseModImmed,Button2Mask},
    {"Button3",	0,	ParseModImmed,Button3Mask},
    {"Button4",	0,	ParseModImmed,Button4Mask},
    {"Button5",	0,	ParseModImmed,Button5Mask},

    {"Any",	0,	ParseModImmed,AnyModifier},

    {"c",	0,	ParseModImmed,ControlMask},
    {"s",	0,	ParseModImmed,ShiftMask},
    {"l",	0,	ParseModImmed,LockMask},
};

static NameValueRec buttonNames[] = {
    {"Button1",	0,	Button1},
    {"Button2", 0,	Button2},
    {"Button3", 0,	Button3},
    {"Button4", 0,	Button4},
    {"Button5", 0,	Button5},
    {NULL, NULLQUARK, 0},
};

static NameValueRec motionDetails[] = {
    {"Normal",		0,	NotifyNormal},
    {"Hint",		0,	NotifyHint},
    {NULL, NULLQUARK, 0},
};

static NameValueRec notifyModes[] = {
    {"Normal",		0,	NotifyNormal},
    {"Grab",		0,	NotifyGrab},
    {"Ungrab",		0,	NotifyUngrab},
    {"WhileGrabbed",    0,	NotifyWhileGrabbed},
    {NULL, NULLQUARK, 0},
};

static NameValueRec mappingNotify[] = {
    {"Modifier",	0,	MappingModifier},
    {"Keyboard",	0,	MappingKeyboard},
    {"Pointer",	0,	MappingPointer},
    {NULL, NULLQUARK, 0},
};

static String ParseKeySym();
static String ParseKeyAndModifiers();
static String ParseTable();
static String ParseImmed();
static String ParseAddModifier();
static String ParseNone();
static String ParseAtom();

static EventKey	events[] = {

/* Event,   Quark,	Event Type,  Detail Parser,	  Closure */

{"Key",	    NULLQUARK,  KeyPress,    ParseKeySym,	  NULL},
{"Ctrl",    NULLQUARK,  KeyPress,    ParseKeyAndModifiers,(Opaque)ControlMask},
{"Shift",   NULLQUARK,  KeyPress,    ParseKeyAndModifiers,(Opaque)ShiftMask},

{"FocusIn", NULLQUARK,  FocusIn,     ParseTable,	  (Opaque)notifyModes},
{"FocusOut",NULLQUARK,  FocusOut,    ParseTable,	  (Opaque)notifyModes},
};

Private Boolean	TMMatchEvent();

#define ScanFor(str, ch) \
    while ((*(str) != (ch)) && (*(str) != '\0') && (*(str) != '\n')) (str)++

#define ScanNumeric(str)  while ('0' <= *(str) && *(str) <= '9') (str)++

#define ScanAlphanumeric(str) \
    while (('A' <= *(str) && *(str) <= 'Z') || \
           ('a' <= *(str) && *(str) <= 'z') || \
           ('0' <= *(str) && *(str) <= '9')) (str)++

#define ScanWhitespace(str) \
    while (*(str) == ' ' || *(str) == '\t') (str)++


static Boolean initialized = False;


Private void *
_HttMalloc(size)
    int		size;
{
    register void	*ptr = (void *)malloc(size);

    bzero( ptr, size );
    return ptr;
}

Private void *
_HttRealloc( ptr, size )
    void	*ptr;
    int		 size;
{
    if ( !ptr )
	return _HttMalloc( size );
    return realloc( ptr, size );
}

Private void
_HttFree( ptr )
    void	*ptr;
{
    if ( ptr )
	free( ptr );
    return;
}

Private String *
DupParams( old, num )
    String	*old;
    Cardinal	 num;
{
    String	*new;
    int		 i;

    if ( !old )
	return NULL;

    new = (String *)XtMalloc( sizeof(String) * (num + 1) );
    for ( i = 0; i < num; i++ ) {
	new[i] = (String )XtMalloc( strlen(old[i]) + 1 );
	strcpy( new[i], old[i] );
    }
    return new;
}

Private ActionPtr
DupActions( old )
    ActionPtr	old;
{
    ActionPtr	top = NULL;
    ActionPtr	last = NULL;

    if ( !old )
	return NULL;

    for ( ; old != NULL; old = old->next ) {
	if ( !top ) {
	    top = last = XtNew( ActionRec );
	} else {
	    last->next = XtNew( ActionRec );
	    last = last->next;
	}
	last->idx = old->idx;
	last->params = DupParams( old->params, old->num_params );
	last->num_params = old->num_params;
    }
    return top;
}

Private EventSeqPtr
DupLinkedEventSeq( old )
    EventSeqPtr	old;
{
    EventSeqPtr	top = NULL;
    EventSeqPtr	next_seq = NULL;

    for ( ; old != NULL; old = old->next ) {
	if ( !top ) {
	    top = next_seq = XtNew( EventSeqRec );
	} else {
	    next_seq->next = XtNew( EventSeqRec );
	    next_seq = next_seq->next;
	}
	bcopy( old, next_seq, sizeof(EventSeqRec) );
	next_seq->actions = DupActions( old->actions );
    }
    return top;
}

Private EventSeqPtr
DupAllEventSeq( old )
    EventSeqPtr	old;
{
    EventSeqPtr	top = NULL;
    EventSeqPtr	next_seq = NULL;

    if ( !old )
	return NULL;

    for ( ; old != NULL; old = old->next_level ) {
	if ( !top ) {
	    top = next_seq = DupLinkedEventSeq( old );
	} else {
	    next_seq->next_level = DupLinkedEventSeq( old );
	    next_seq = next_seq->next_level;
	}
    }
    return top;
}

Private TMParseStateTree
DupParseTree( old )
    TMParseStateTree	old;
{
    TMParseStateTree	new;

    if ( !old )
	return NULL;
    new = XtNew( TMParseStateTreeRec );
    bcopy( old, new, sizeof(TMParseStateTreeRec) );
    new->eventSeq = DupAllEventSeq( old->eventSeq );
    return new;
}

Private HttTranslations
DupTranslations( old )
    HttTranslations	old;
{
    HttTranslations	new;

    if ( !old )
	return NULL;
    new = XtNew( TranslationData );
    bcopy( old, new, sizeof(HttTranslations) );
    new->parseTree = DupParseTree( old->parseTree );
    return new;
}

Private void
freeActionList( list )
    ActionList	list;
{
    ActionList	next;

    for ( ; list; list = next ) {
	next = list->next;
	XtFree( list );
    }
    return;
}

Private void
freeActionRec( action )
    ActionPtr	action;
{
    int		i;

    if ( !action )
	return;
    if ( action->next )
	freeActionRec( action->next );
    for ( i = 0; i < action->num_params; i++ )
	XtFree( action->params[i] );
    XtFree( action->params );
    XtFree( action );
    return;
}

Private void
freeEventSeq( eventSeq )
    EventSeqPtr	 eventSeq;
{
    ActionPtr	action, nAction;

    if ( !eventSeq )
	return;
    if ( eventSeq->next_level )
	freeEventSeq( eventSeq->next_level );
    if ( eventSeq->next )
	freeEventSeq( eventSeq->next );
    freeActionRec( eventSeq->actions );
    XtFree( eventSeq );
    return;
}

Private void
freeTranslations( translations )
    HttTranslations	translations;
{
    if ( !translations )
	return;
    if ( translations->parseTree ) {
	freeEventSeq( translations->parseTree->eventSeq );
	XtFree( translations->parseTree );
    }
    XtFree( translations );
    return;
}

Private void
FreeEventSeq(eventSeq)
    EventSeqPtr eventSeq;
{
    register EventSeqPtr evs = eventSeq;

    while (evs != NULL) {
	evs->state = (StatePtr) evs;
	if (evs->next != NULL
	    && evs->next->state == (StatePtr) evs->next)
	    evs->next = NULL;
	evs = evs->next;
    }

    evs = eventSeq;
    while (evs != NULL) {
	register EventPtr event = evs;
	evs = evs->next;
	if (evs == event) evs = NULL;
	XtFree((char *)event);
    }
}

Private int
OrderEvents(a, b)
    EventKey	*a, *b;
{
    return ((a->signature < b->signature) ? -1 : 1);
}

Private void
CompileNameValueTable( table )
    NameValueTable	table;
{
    register int	i;

    for ( i = 0; table[i].name; i++ )
	table[i].signature = XrmPermStringToQuark( table[i].name );
    return;
}

Private void
Compile_XtEventTable(table, count)
    EventKeys	table;
    Cardinal	count;
{
    register int i;
    register EventKeys entry = table;

    for (i=count; --i >= 0; entry++)
	entry->signature = XrmPermStringToQuark(entry->event);
    qsort(table, count, sizeof(EventKey), OrderEvents);
}

Private int
OrderModifiers(a, b)
    ModifierRec *a, *b;
{
    return ((a->signature < b->signature) ? -1 : 1);
}

Private void
Compile_XtModifierTable(table, count)
    ModifierKeys table;
    Cardinal count;
{
    register int i;
    register ModifierKeys entry = table;

    for (i=count; --i >= 0; entry++)
	entry->signature = XrmPermStringToQuark(entry->name);
    qsort(table, count, sizeof(ModifierRec), OrderModifiers);
}

Private String
PanicModeRecovery(str)
    String str;
{
     ScanFor(str,'\n');
     if (*str == '\n') str++;
     return str;

}


Private
Syntax(str,str1)
    String str,str1;
{
    Cardinal numChars;
    Cardinal num_params = 1;
    String params[1];
    char message[1000];

    (void)strcpy(message,str);
    numChars = strlen(message);
    (void) strcpy(&message[numChars], str1);
    numChars += strlen(str1);
    message[numChars] = '\0';
    params[0] = (String )message;
    XtWarningMsg(XtNtranslationParseError,"parseError",XtCXtToolkitError,
		 "translation table syntax error: %s",params,&num_params);
}

Private Cardinal
LookupTMEventType(eventStr,error)
  String eventStr;
  Boolean *error;
{
    register int   i, left, right;
    register XrmQuark	signature;
    static int 	previous = 0;

    if ((signature = StringToQuark(eventStr)) == events[previous].signature)
	return (Cardinal) previous;

    left = 0;
    right = XtNumber(events) - 1;
    while (left <= right) {
	i = (left + right) >> 1;
	if (signature < events[i].signature)
	    right = i - 1;
	else if (signature > events[i].signature)
	    left = i + 1;
	else {
	    previous = i;
	    return (Cardinal) i;
	}
    }

    Syntax("Unknown event type :  ",eventStr);
    *error = TRUE;
    return (Cardinal) i;
}

Private Boolean
_XtLookupTableSym( table, name, valueP )
    NameValueTable	 table;
    String		 name;
    Value		*valueP;
{
    register int	i;
    register XrmQuark	signature = StringToQuark( name );

    for ( i = 0; table[i].name != NULL; i++ )
	if ( table[i].signature == signature ) {
	    *valueP = table[i].value;
	    return True;
	}

    return False;
}

static void StoreLateBindings(keysymL,notL,keysymR,notR,lateBindings)
    KeySym  keysymL;
    Boolean notL;
    KeySym keysymR;
    Boolean notR;
    LateBindingsPtr* lateBindings;
{
    LateBindingsPtr temp;
    Boolean pair = False;
    unsigned long count,number;
    if (lateBindings != NULL){
        temp = *lateBindings;
        if (temp != NULL) {
            for (count = 0; temp[count].keysym; count++){/*EMPTY*/}
        }
        else count = 0;
        if (! keysymR){
             number = 1;pair = False;
        } else{
             number = 2;pair = True;
        }
          
        temp = (LateBindingsPtr)XtRealloc((char *)temp,
            (unsigned)((count+number+1) * sizeof(LateBindings)) );
        *lateBindings = temp;
        temp[count].knot = notL;
        temp[count].pair = pair;
	/*
	if (count == 0)
	    temp[count].ref_count = 1;
	*/
        temp[count++].keysym = keysymL;
        if (keysymR){
            temp[count].knot = notR;
            temp[count].pair = False;
            temp[count++].keysym = keysymR;
        }
        temp[count].knot = False;
        temp[count].keysym = 0;
    }
} 

Private Boolean
_XtLookupModifier(name,lateBindings,notFlag,valueP,constMask)
    String name;
    LateBindingsPtr* lateBindings;
    Boolean notFlag;
    Value *valueP;
    Bool constMask;
{
   register int i, left, right;
   register XrmQuark signature = XrmStringToQuark(name);
   static int previous = 0;
   
   if (signature == modifiers[previous].signature) {
       if (constMask)  *valueP = modifiers[previous].value;
       else /* if (modifiers[previous].modifierParseProc) always true */
	   (*modifiers[previous].modifierParseProc)
	      (name, modifiers[previous].value, lateBindings, notFlag, valueP);
       return True;
   }

   left = 0;
   right = XtNumber(modifiers) - 1;
   while (left <= right) {
       i = (left + right) >> 1;
       if (signature < modifiers[i].signature)
	   right = i - 1;
       else if (signature > modifiers[i].signature)
	   left = i + 1;
       else {
	   previous = i;
	   if (constMask)  *valueP = modifiers[i].value;
	   else /* if (modifiers[i].modifierParseProc) always true */
	       (*modifiers[i].modifierParseProc)
		   (name, modifiers[i].value, lateBindings, notFlag, valueP);
	   return True;
       }
   }
   return False;
}


Private String
ScanIdent(str)
    register String str;
{
    ScanAlphanumeric(str);
    while (
	   ('A' <= *str && *str <= 'Z')
	|| ('a' <= *str && *str <= 'z')
	|| ('0' <= *str && *str <= '9')
	|| (*str == '-')
	|| (*str == '_')
	|| (*str == '$')
	) str++;
    return str;
}

static unsigned long StrToHex(str)
    String str;
{
    register char   c;
    register unsigned long    val = 0;

    while (c = *str) {
	if ('0' <= c && c <= '9') val = val*16+c-'0';
	else if ('a' <= c && c <= 'z') val = val*16+c-'a'+10;
	else if ('A' <= c && c <= 'Z') val = val*16+c-'A'+10;
	else return 0;
	str++;
    }

    return val;
}

static unsigned long StrToOct(str)
    String str;
{
    register char c;
    register unsigned long  val = 0;

    while (c = *str) {
        if ('0' <= c && c <= '7') val = val*8+c-'0'; else return 0;
	str++;
    }

    return val;
}

static unsigned long StrToNum(str)
    String str;
{
    register char c;
    register unsigned long val = 0;

    if (*str == '0') {
	str++;
	if (*str == 'x' || *str == 'X') return StrToHex(++str);
	else return StrToOct(str);
    }

    while (c = *str) {
	if ('0' <= c && c <= '9') val = val*10+c-'0';
	else return 0;
	str++;
    }

    return val;
}

static KeySym StringToKeySym(str, error)
    String str;
    Boolean *error;
{
    KeySym k;

    if (str == NULL || *str == '\0') return (KeySym) 0;

#ifndef NOTASCII
    /* special case single character ASCII, for speed */
    if (*(str+1) == '\0') {
	if (' ' <= *str && *str <= '~') return XK_space + (*str - ' ');
    }
#endif

    if ('0' <= *str && *str <= '9') return (KeySym) StrToNum(str);
    k = XStringToKeysym(str);
    if (k != NoSymbol) return k;

#ifdef NOTASCII
    /* fall-back case to preserve backwards compatibility; no-one
     * should be relying upon this!
     */
    if (*(str+1) == '\0') return (KeySym) *str;
#endif

    Syntax("Unknown keysym name: ", str);
    *error = True;
    return NoSymbol;
}

/* ARGSUSED */
Private void
ParseModImmed(name,value,lateBindings,notFlag,valueP)
    String name;
    Value value;
    LateBindingsPtr* lateBindings;
    Boolean notFlag;
    Value* valueP;
{
    *valueP = value;
}

/* ARGSUSED */
Private void
ParseModSym (name,value,lateBindings,notFlag,valueP)
 /* is only valid with keysyms that have an _L and _R in their name;
  * and ignores keysym lookup errors (i.e. assumes only valid keysyms)
  */
    String name;
    Value value;
    LateBindingsPtr* lateBindings;
    Boolean notFlag;
    Value* valueP;
{
    register KeySym keysymL = (KeySym)value;
    register KeySym keysymR = keysymL + 1; /* valid for supported keysyms */
    StoreLateBindings(keysymL,notFlag,keysymR,notFlag,lateBindings);
    *valueP = 0;
}

Private void
TranslateKeycode( display, eventCode, modifier, r_modifier, r_keysym )
    Display	*display;
    TMLongCard	 eventCode;
    TMLongCard	 modifier;
    TMLongCard	*r_modifier;
    KeySym	*r_keysym;
{
    KeySym	 u_keysym;
    KeySym	 l_keysym;

    if ( !modifier ) {
	*r_keysym = l_keysym = XKeycodeToKeysym( display, eventCode, 0 );
    } else {
	u_keysym = XKeycodeToKeysym( display, eventCode, 1 );
	if ( u_keysym == NoSymbol )
	    u_keysym = XKeycodeToKeysym( display, eventCode, 0 );
	*r_keysym = u_keysym;
    }
    return;
}

#ifdef sparc
/*
 * The stupid optimizer in SunOS 4.0.3 and below generates bogus code that
 * causes the value of the most recently used variable to be returned instead
 * of the value passed in.
 */
Private String stupid_optimizer_kludge;
#define BROKEN_OPTIMIZER_HACK(val) stupid_optimizer_kludge = (val)
#else
#define BROKEN_OPTIMIZER_HACK(val) val
#endif

Private Boolean
_XtMatchUsingStandardMods( tmSeq, curSeq )
    EventSeqPtr		tmSeq;
    TMEventSeqPtr	curSeq;
{
    Event	*typeMatch = &tmSeq->event;
    Event	*curEvent  = &curSeq->tmEvent.event;
    KeySym	 r_keysym;
    TMLongCard	 r_modifier;
    TMLongCard	 modifier;
    TMLongCard	 temp = 0;
    TMLongCard	 tempMask = 0;

    if ( typeMatch->modifiers &
		typeMatch->modifierMask & (ShiftMask|LockMask) )
	modifier = curEvent->modifiers &
			typeMatch->modifierMask & (ShiftMask|LockMask);
    else
	modifier = 0L;
    TranslateKeycode( curSeq->tmEvent.xev->xkey.display,
			curEvent->eventCode, modifier, &r_modifier, &r_keysym );

    if ( ( typeMatch->eventCode & typeMatch->eventCodeMask ) ==
	 ( r_keysym & typeMatch->eventCodeMask ) ) {
	temp |= typeMatch->modifiers;
	tempMask |= typeMatch->modifierMask;
	if ( ( temp & tempMask ) ==
	     ( curEvent->modifiers & tempMask ) )
	    return True;
    }
    return False;
}

Private Boolean
_XtRegularMatch( tmSeq, curSeq )
    TMEventSeqPtr	tmSeq;
    TMEventSeqPtr	curSeq;
{
    Event	*tmEvent  = &tmSeq->tmEvent.event;
    Event	*curEvent = &curSeq->tmEvent.event;

    if ( ( tmEvent->modifiers & tmEvent->modifierMask ) == (curEvent->modifiers & tmEvent->modifierMask ) &&
	 ( tmEvent->eventCode & tmEvent->eventCodeMask ) == ( curEvent->eventCode & tmEvent->eventCodeMask ) )
	return True;
    return False;
}

Private String
ParseQuotedStringEvent(str, event,error)
    register String str;
    register EventPtr event;
    Boolean *error;
{
    register int j;

    Value ctrlMask;
    Value metaMask;
    Value shiftMask;
    Value lockMask;
    register char	c;
    char	s[2];
    (void) _XtLookupModifier("Ctrl",(LateBindingsPtr*)NULL,
                 False,(Value *) &ctrlMask,True);
    (void) _XtLookupModifier("Meta", &event->event.lateModifiers,
                 False,(Value *) &metaMask,False);
    (void) _XtLookupModifier("Shift",(LateBindingsPtr*)NULL,
                 False,(Value *) &shiftMask,True);
    (void) _XtLookupModifier("Lock",(LateBindingsPtr*)NULL,
                 False,(Value *) &lockMask,True);
    for (j=0; j < 2; j++)
	if (*str=='^' && !(event->event.modifiers & ctrlMask)) {
	    str++;
	    event->event.modifiers |= ctrlMask;
	} else if (*str == '$' && !(event->event.modifiers & metaMask)) {
	    str++;
	    event->event.modifiers |= metaMask;
	} else if (*str == '\\') {
	    str++;
	    c = *str;
	    if (*str != '\0' && *str != '\n') str++;
	    break;
	} else {
	    c = *str;
	    if (*str != '\0' && *str != '\n') str++;
	    break;
	}
    event->event.eventType = KeyPress;
    s[0] = c;
    s[1] = '\0';
    event->event.eventCode = StringToKeySym(s, error);
    if (*error) return PanicModeRecovery(str);
    event->event.eventCodeMask = ~0L;
    event->event.modifierMask = ~( shiftMask | lockMask );
    event->event.matchEvent = _XtMatchUsingStandardMods;
    event->event.standard = True;

    return str;
}

static String FetchModifierToken(str,modStr)
    String str,modStr;
{
    String start = str;
    String metaString = "Meta";
    String ctrlString = "Ctrl";
    if (*str == '$') {
        strcpy(modStr,metaString);
        str++;
        return str;
    }
    if (*str == '^') {
        strcpy(modStr,ctrlString);
        str++;
        return str;
    }
    str = ScanIdent(str);
    if (start != str) {
	 bcopy(start, modStr, str-start);
          modStr[str-start] = '\0';
          return str;
    }
    return str;
}        

static _XtParseKeysymMod(name,lateBindings,notFlag,valueP,error)
    String name;
    LateBindingsPtr* lateBindings;
    Boolean notFlag;
    Value *valueP;
    Boolean *error;
{
    KeySym keySym;
    keySym = StringToKeySym(name, error);
    *valueP = 0;
    if (keySym != NoSymbol) {
        StoreLateBindings(keySym,notFlag,(KeySym) NULL,False,lateBindings);
    }
}

static String ParseModifiers(str, event, error)
    register String str;
    EventPtr event;
    Boolean* error;
{
    register String start;
    Boolean notFlag, exclusive, keysymAsMod;
    Value maskBit;
    char modStr[100];
 
    ScanWhitespace(str);
    start = str;
    str = FetchModifierToken(str,modStr);
    exclusive = False;
    if (start != str) {
          if (_XtLookupModifier(modStr,(LateBindingsPtr *) NULL,
		  False,&maskBit,True)) {
	      if (maskBit== None) {
		  event->event.modifierMask = ~0;
		  event->event.modifiers = 0;
		  ScanWhitespace(str);
		  return str;
	      }
	      if (maskBit == AnyModifier) {/*backward compatability*/
		  event->event.modifierMask = 0;
		  event->event.modifiers = 0;
		  ScanWhitespace(str);
		  return str;
	      }
	  }
	  str = start; /*if plain modifier, reset to beginning */
    }
    else while (*str == '!' || *str == ':') {
        if (*str == '!') {
             exclusive = True;
             str++;
             ScanWhitespace(str);
        }
        if (*str == ':') {
             event->event.standard = True;
             str++;
             ScanWhitespace(str);
        }
    }
   
    while (*str != '<') {
        if (*str == '~') {
             notFlag = True;
             str++;
          } else 
              notFlag = False;
        if (*str == '@') {
            keysymAsMod = True;
            str++;
        }
        else keysymAsMod = False;
	start = str;
        str = FetchModifierToken(str,modStr);
        if (start == str) {
            Syntax("Modifier or '<' expected","");
            *error = True;
            return PanicModeRecovery(str);
        }
         if (keysymAsMod) {
             _XtParseKeysymMod(modStr,&event->event.lateModifiers,
			       notFlag,&maskBit, error);
	     if (*error)
                 return PanicModeRecovery(str);

         } else
  	     if (!_XtLookupModifier( modStr,
	   &event->event.lateModifiers, notFlag, &maskBit,False)) {
	         Syntax("Unknown modifier name:  ",modStr);
                 *error = True;
                 return PanicModeRecovery(str);
             }
        event->event.modifierMask |= maskBit;
	if (notFlag)
	    event->event.modifiers &= ~maskBit;
	else
	    event->event.modifiers |= maskBit;
        ScanWhitespace(str);
    }
    if (exclusive)
	event->event.modifierMask = ~0;
    return str;
}

Private String
ParseXtEventType( str, event, tmEventP, error )
    register String	 str;
    EventPtr		 event;
    Cardinal		*tmEventP;
    Boolean		*error;
{
    String	start = str;
    char	eventTypeStr[100];

    ScanAlphanumeric( str );
    bcopy( start, eventTypeStr, str - start );
    eventTypeStr[str - start] = '\0';
    *tmEventP = LookupTMEventType( eventTypeStr, error );
    if ( *error )
	return PanicModeRecovery( str );
    event->event.eventType = events[*tmEventP].eventType;
    return str;
}

Private String ParseEvent( str, e, error )
    register String	 str;
    EventPtr		 e;
    Boolean		*error;
{
    Cardinal		 tmEvent;

    str = ParseModifiers( str, e, error );
    if ( *error ) return str;
    if ( *str != '<' ) {
	*error = True;
	return PanicModeRecovery( str );
    } else str ++;
    str = ParseXtEventType( str, e, &tmEvent, error );
    if ( *error ) return str;
    if ( *str != '>' ) {
	*error = True;
	return PanicModeRecovery( str );
    } else str ++;
    str = (*(events[tmEvent].parseDetail))(
		str, events[tmEvent].closure, e, error );
    if ( *error ) return str;
    return str;
}

Private String
ParseKeyAndModifiers( str, closure, event, error )
    register String str;
    Opaque closure;
    EventPtr event;
    Boolean* error;
{
    str = ParseKeySym( str, closure, event, error );

    event->event.modifiers |= (unsigned long)closure;
    event->event.modifierMask |= (unsigned long)closure;

    return str;
}

/*ARGSUSED*/
Private String
ParseKeySym(str, closure, event, error)
    register String str;
    Opaque closure;
    EventPtr event;
    Boolean* error;
{
    char *start;
    char keySymName[100];

    ScanWhitespace(str);

    if (*str == '\\') {
	str++;
	keySymName[0] = *str;
	if (*str != '\0' && *str != '\n') str++;
	keySymName[1] = '\0';
	event->event.eventCode = StringToKeySym(keySymName, error);
	event->event.eventCodeMask = ~0L;
    } else if (*str == ',' || *str == ':' ||
             /* allow leftparen to be single char symbol,
              * for backwards compatibility
              */
             (*str == '(' && *(str+1) >= '0' && *(str+1) <= '9')) {
	/* no detail */
	event->event.eventCode = 0L;
        event->event.eventCodeMask = 0L;
    } else {
	start = str;
	while (
		*str != ','
		&& *str != ':'
		&& *str != ' '
		&& *str != '\t'
                && *str != '\n'
                && (*str != '(' || *(str+1) <= '0' || *(str+1) >= '9')
		&& *str != '\0') str++;
	bcopy(start, keySymName, str-start);
	keySymName[str-start] = '\0';
	event->event.eventCode = StringToKeySym(keySymName, error);
	event->event.eventCodeMask = ~0L;
    }
    if (*error) {
	if (keySymName[0] == '<') {
	    /* special case for common error */
	    XtWarningMsg(XtNtranslationParseError, "missingComma",
			 XtCXtToolkitError,
		     "... possibly due to missing ',' in event sequence.",
		     (String*)NULL, (Cardinal*)NULL);
	}
	return PanicModeRecovery(str);
    }
    event->event.matchEvent = _XtMatchUsingStandardMods;

    return str;
}

Private String
ParseTable( str, closure, event, error )
    register String	 str;
    Opaque		 closure;
    EventPtr		 event;
    Boolean		*error;
{
    register String	start = str;
    char		tableSymName[100];

    event->event.eventCode = 0L;
    ScanAlphanumeric( str );
    if ( str == start ) { event->event.eventCodeMask = 0L; return str; }
    if ( str - start >= 99 ) {
	Syntax("Invalid Detail Type (string is too long).", "" );
	*error = True;
	return str;
    }
    bcopy( start, tableSymName, str - start );
    tableSymName[str - start] = '\0';
    if ( ! _XtLookupTableSym( (NameValueTable)closure, tableSymName,
		(Value *)&event->event.eventCode ) ) {
	Syntax( "Unknown Detail Type:  ", tableSymName );
	*error = True;
	return PanicModeRecovery( str );
    }
    event->event.eventCode = ~0L;

    return str;
}

/***********************************************************************
 * ParseEventSeq
 * Parses the left hand side of a translation table production
 * up to, and consuming the ":".
 * Takes a pointer to a char* (where to start parsing) and returns an
 * event seq (in a passed in variable), having updated the String 
 **********************************************************************/

Private String
ParseEventSeq(str, eventSeqP, actionsP,error)
    register String str;
    EventSeqPtr *eventSeqP;
    ActionPtr	**actionsP;
    Boolean *error;
{
    EventSeqPtr *nextEvent = eventSeqP;

    *eventSeqP = NULL;

    while ( *str != '\0' && *str != '\n') {
	static Event	nullEvent =
             {0, 0, 0L, 0, 0L, 0L, _XtRegularMatch, False};
	EventPtr	event;

	ScanWhitespace(str);

	if (*str == '"') {
	    str++;
	    while (*str != '"' && *str != '\0' && *str != '\n') {
                event = XtNew(EventRec);
                event->event = nullEvent;
                event->state = /* (StatePtr) -1 */ NULL;
                event->next = NULL;
                event->actions = NULL;
		str = ParseQuotedStringEvent(str, event,error);
		if (*error) {
		    XtWarningMsg(XtNtranslationParseError, "nonLatin1",
			XtCXtToolkitError,
			"... probably due to non-Latin1 character in quoted string",
			(String*)NULL, (Cardinal*)NULL);
		    return PanicModeRecovery(str);
		}
		*nextEvent = event;
		*actionsP = &event->actions;
		nextEvent = &event->next;
	    }
	    if (*str != '"') {
                Syntax("Missing '\"'.","");
                *error = True;
                return PanicModeRecovery(str);
             }
             else str++;
	} else {
            event = XtNew(EventRec);
            event->event = nullEvent;
            event->state = /* (StatePtr) -1 */ NULL;
            event->next = NULL;
            event->actions = NULL;

	    str = ParseEvent( str, event, error );
	    if ( *error ) return str;
	    *nextEvent = event;
	    *actionsP = &event->actions;
	    nextEvent = &event->next;
	}
	ScanWhitespace(str);
        if (*str == ':') break;
        else {
            if (*str != ',') {
                Syntax("',' or ':' expected while parsing event sequence.","");
                *error = True;
                return PanicModeRecovery(str);
	    } else str++;
        }
    }

    if (*str != ':') {
        Syntax("Missing ':'after event sequence.",""); 
        *error = True;
        return PanicModeRecovery(str);
    } else str++;

    return str;
}


Private String
ParseActionProc(str, actionProcNameP, error)
    register String str;
    XrmQuark *actionProcNameP;
    Boolean *error;
{
    register String start = str;
    char procName[200];

    str = ScanIdent(str);
    if (str-start >= 199) {
	Syntax("Action procedure name is longer than 199 chars","");
	*error = True;
	return str;
    }
    bcopy(start, procName, str-start);
    procName[str-start] = '\0';
    *actionProcNameP = XrmStringToQuark( procName );
    return str;
}


Private String
ParseString(str, strP)
    register String str;
    String *strP;
{
    register String start;

    if (*str == '"') {
	register unsigned prev_len, len;
	str++;
	start = str;
	*strP = NULL;
	prev_len = 0;

	while (*str != '"' && *str != '\0') {
	    /* \"  produces double quote embedded in a quoted parameter
	     * \\" produces backslash as last character of a quoted parameter
	     */
	    if (*str == '\\' && 
		(*(str+1) == '"' || (*(str+1) == '\\' && *(str+2) == '"'))) {
		len = prev_len + (str-start+2);
		*strP = XtRealloc(*strP, len);
		bcopy(start, *strP + prev_len, str-start);
		prev_len = len-1;
		str++;
		(*strP)[prev_len-1] = *str;
		(*strP)[prev_len] = '\0';
		start = str+1;
	    } 
	    str++;
	}
	len = prev_len + (str-start+1);
	*strP = XtRealloc(*strP, len);
	bcopy(start, *strP + prev_len, str-start);
	(*strP)[len-1] = '\0';
	if (*str == '"') str++; else
            XtWarningMsg(XtNtranslationParseError,"parseString",
                      XtCXtToolkitError,"Missing '\"'.",
		      (String *)NULL, (Cardinal *)NULL);
    } else {
	/* scan non-quoted string, stop on whitespace, ',' or ')' */
	start = str;
	while (*str != ' '
		&& *str != '\t'
		&& *str != ','
		&& *str != ')'
                && *str != '\n'
		&& *str != '\0') str++;
	*strP = XtMalloc((unsigned)(str-start+1));
	bcopy(start, *strP, str-start);
	(*strP)[str-start] = '\0';
    }
    return str;
}


Private String
ParseParamSeq(str, paramSeqP, paramNumP)
    register String str;
    String **paramSeqP;
    Cardinal *paramNumP;
{
#define ALLOCATE_LOCAL(size)	malloc(size)
#define DEALLOCATE_LOCAL(ptr)	free(ptr)

    typedef struct _ParamRec *ParamPtr;
    typedef struct _ParamRec {
	ParamPtr next;
	String	param;
    } ParamRec;

    ParamPtr params = NULL;
    register Cardinal num_params = 0;
    register Cardinal i;

    ScanWhitespace(str);
    while (*str != ')' && *str != '\0' && *str != '\n') {
	String newStr;
	str = ParseString(str, &newStr);
	if (newStr != NULL) {
	    ParamPtr temp = (ParamRec*)
		ALLOCATE_LOCAL( (unsigned)sizeof(ParamRec) );

	    num_params++;
	    temp->next = params;
	    params = temp;
	    temp->param = newStr;
	    ScanWhitespace(str);
	    if (*str == ',') {
		str++;
		ScanWhitespace(str);
	    }
	}
    }

    if (num_params != 0) {
	String *paramP = (String *)
		XtMalloc( (unsigned)(num_params+1) * sizeof(String) );
	*paramSeqP = paramP;
	*paramNumP = num_params;
	paramP += num_params; /* list is LIFO right now */
	*paramP-- = NULL;
	for (i=0; i < num_params; i++) {
	    ParamPtr next = params->next;
	    *paramP-- = params->param;
	    DEALLOCATE_LOCAL( (char *)params );
	    params = next;
	}
    } else {
	*paramSeqP = NULL;
	*paramNumP = 0;
    }

    return str;
}

Private String
ParseAction(str, actionP, quarkP, error)
    String str;
    ActionPtr actionP;
    XrmQuark* quarkP;
    Boolean* error;
{
    str = ParseActionProc(str, quarkP, error);
    if (*error) return str;

    if (*str == '(') {
	str++;
	str = ParseParamSeq(str, &actionP->params, &actionP->num_params);
    } else {
        Syntax("Missing '(' while parsing action sequence",""); 
        *error = True;
        return str;
    }
    if (*str == ')') str++;
    else{
        Syntax("Missing ')' while parsing action sequence","");
        *error = True;
        return str;
    }
    return str;
}


Private String
ParseActionSeq(tc, parseTree, str, actionsP, error)
    HttTCid		tc;
    TMParseStateTree   	parseTree;
    String 		str;
    ActionPtr 		*actionsP;
    Boolean		*error;
{
    ActionPtr *nextActionP = actionsP;

    *actionsP = NULL;
    while (*str != '\0' && *str != '\n') {
	register ActionPtr	action;
	XrmQuark quark;

	action = XtNew(ActionRec);
        action->params = NULL;
        action->num_params = 0;
        action->next = NULL;

	str = ParseAction(str, action, &quark, error);
	if (*error) return PanicModeRecovery(str);

	action->idx = _XtGetQuarkIndex(tc, quark);
	ScanWhitespace(str);
	*nextActionP = action;
	nextActionP = &action->next;
    }
    if (*str == '\n') str++;
    ScanWhitespace(str);
    return str;
}


Private void
ShowProduction(currentProduction)
  String currentProduction;
{
    Cardinal num_params = 1;
    String params[1];
    int len = 499;
    char *eol, production[500];

    eol = (char *)strchr(currentProduction, '\n');
    if (eol) len = MIN(499, eol - currentProduction);
    bcopy(currentProduction, production, len);
    production[len] = '\0';

    params[0] = production;
    XtWarningMsg(XtNtranslationParseError, "showLine", XtCXtToolkitError,
		 "... found while parsing '%s'", params, &num_params);
}

Private void
_XtAddEventSeqToStateTree( eventSeqPtr, parseTree )
    EventSeqPtr		eventSeqPtr;
    TMParseStateTree	parseTree;
{
    EventSeqPtr		esP = parseTree->eventSeq;
    EventSeqPtr		lesP = parseTree->eventSeq;

    if ( !esP ) {
	parseTree->eventSeq = eventSeqPtr;
	return;
    }
    for ( ; esP; lesP = esP, esP = esP->next_level );
    lesP->next_level = eventSeqPtr;
    return;
}

/***********************************************************************
 * ParseTranslationTableProduction
 * Parses one line of event bindings.
 ***********************************************************************/

Private String
ParseTranslationTableProduction(tc, parseTree, str)
    HttTCid		 tc;
    TMParseStateTree	 parseTree;
    register String	 str;
{
    EventSeqPtr	eventSeq = NULL;
    ActionPtr	*actionsP;
    Boolean error = False;
    String	production = str;

    str = ParseEventSeq(str, &eventSeq, &actionsP,&error);
    if (error == True) {
	ShowProduction(production);
        FreeEventSeq(eventSeq);
        return (str);
    }
    ScanWhitespace(str);
    str = ParseActionSeq(tc, parseTree, str, actionsP, &error);
    if (error) {
	ShowProduction(production);
        FreeEventSeq(eventSeq);
        return (str);
    }

    _XtAddEventSeqToStateTree(eventSeq, parseTree);
    return (str);
}

Private String
CheckForPoundSign(str, defaultOp, actualOpRtn)
    String str;
    _XtTranslateOp defaultOp;
    _XtTranslateOp *actualOpRtn;
{
    String start;
    char operation[20];
    _XtTranslateOp opType;
    
    opType = defaultOp;
    ScanWhitespace(str);
    if (*str == '#') {
	int len;
	str++;
	start = str;
	str = ScanIdent(str);
	len = MIN(19, str-start);
	bcopy(start, operation, len);
	operation[len] = '\0';
	if (!strcmp(operation,"replace"))
	  opType = XtTableReplace;
	else if (!strcmp(operation,"augment"))
	  opType = XtTableAugment;
	else if (!strcmp(operation,"override"))
	  opType = XtTableOverride;
	ScanWhitespace(str);
	if (*str == '\n') {
	    str++;
	    ScanWhitespace(str);
	}
    }
    *actualOpRtn = opType;
    return str;
}

Private HttTranslations
ParseTranslationTable( tc, source, isAccelerator, defaultOp )
    HttTCid		tc;
    String 		source;
    Boolean		isAccelerator;
    _XtTranslateOp	defaultOp;
{
    HttTranslations		 xlations;
    _XtTranslateOp		 actualOp;
    TMParseStateTreeRec		*parseTree;

    if (source == NULL)
      return NULL;

    source = CheckForPoundSign(source, defaultOp, &actualOp);

    parseTree = XtNew( TMParseStateTreeRec );
    tc->isStackQuarks = False;

    while (source != NULL && *source != '\0') {
	source =  ParseTranslationTableProduction( tc, parseTree, source );
    }

    xlations = XtNew( TranslationData );
    xlations->parseTree = parseTree;
    xlations->operation = actualOp;

    return xlations;
}

Private void
XEventToTMEvent( tc, event, tmEvent )
    HttTCid		 tc;
    register XEvent	*event;
    register TMEventPtr	 tmEvent;
{
#define sKana		4

    XComposeStatus	 NotSupported;
    char		 s[8];
    KeySym		 keysym;

    tmEvent->xev = XtNew( XEvent );
    tmEvent->event.eventCodeMask = 0;
    tmEvent->event.modifierMask = 0;
    tmEvent->event.eventType = event->type;
    tmEvent->event.lateModifiers = NULL;
    tmEvent->event.matchEvent = NULL;
    tmEvent->event.standard = False;

    switch ( event->type ) {

    case KeyPress:
    case KeyRelease:
	bcopy( event, tmEvent->xev, sizeof(XKeyEvent) );
	tmEvent->event.eventCode = event->xkey.keycode;
	tmEvent->event.modifiers = event->xkey.state;
	break;

    case FocusIn:
    case FocusOut:
	bcopy( event, tmEvent->xev, sizeof(XFocusChangeEvent) );
	tmEvent->event.eventCode = event->xfocus.mode;
	tmEvent->event.modifiers = 0;
	break;

    default:
	bcopy( event, tmEvent->xev, sizeof(XEvent) );
	tmEvent->event.eventCode = 0;
	tmEvent->event.modifiers = 0;
	break;
    }
    return;
}


Private HttTranslations
ParseTranslation( tc, source )
    HttTCid	 tc;
    String	 source;
{
    HttTranslations	Xlations;
    return ParseTranslationTable( tc, source, False, XtTableReplace );
}

#ifdef RDB_SUPPORT
#define	HttDefaultXrmDatabase()		(tc->sc->If->im->htt->app_defaults)
#define	HttUserXrmDatabase()		(tc->sc->If->im->htt->rdb)
#else
#define HttDefaultXrmDatabase()		XrmGetFileDatabase("httDefault.rdb")
#define HttUserXrmDatabase()		XrmGetFileDatabase("htt.rdb")
#endif

Private HttTranslations
GetDefaultTranslations( tc, name, class )
    HttTCid	 tc;
    String	 name;
    String	 class;
{
    XrmDatabase	 rdb = HttDefaultXrmDatabase();
    String	 type;
    XrmValue	 value;

    if ( XrmGetResource( rdb, name, class, &type, &value ) )
	return ParseTranslation( tc, value.addr );
    return NULL;
}

#ifdef DISPLAY_SUPPORT
#define	HttGetDisplay()	    \
	    (((htt_t *)(((im_t *)(tc->sc->If->im))->htt))->core->display)
#else
#define	HttGetDisplay()	    (display)
#endif

Private HttTranslations
GetXDefaultTranslations( tc, name, class )
    HttTCid	 tc;
    String	 name;
    String	 class;
{
    XrmDatabase	 rdb = XrmGetDatabase(HttGetDisplay());
    String	 type;
    XrmValue	 value;

    if ( XrmGetResource( rdb, name, class, &type, &value ) )
	return ParseTranslation( tc, value.addr );
    return NULL;
}

Private HttTranslations
GetUserTranslations( tc, name, class )
    HttTCid	 tc;
    String	 name;
    String	 class;
{
    XrmDatabase	 rdb = HttUserXrmDatabase();
    String	 type;
    XrmValue	 value;

    if ( XrmGetResource( rdb, name, class, &type, &value ) )
	return ParseTranslation( tc, value.addr );
    return NULL;
}

Private Boolean
CheckEventSeq( oldEventSeq, newEventSeq )
    EventSeqPtr		oldEventSeq;
    EventSeqPtr		newEventSeq;
{
    for ( ;; ) {
	if ( !oldEventSeq && !newEventSeq )
	    break;
	if ( !oldEventSeq || !newEventSeq )
	    return False;
	if ( oldEventSeq->event.modifiers != newEventSeq->event.modifiers ||
	  oldEventSeq->event.modifierMask != newEventSeq->event.modifierMask ||
	  oldEventSeq->event.eventType != newEventSeq->event.eventType ||
	  oldEventSeq->event.eventCode != newEventSeq->event.eventCode ||
	  oldEventSeq->event.eventCodeMask != newEventSeq->event.eventCodeMask )
	    return False;
	oldEventSeq = oldEventSeq->next;
	newEventSeq = newEventSeq->next;
    }
    return True;
}

Private EventSeqPtr
MergeTranslationProc( tc, old, new, opType )
    HttTCid		 tc;
    TMParseStateTree	 old;
    TMParseStateTree	 new;
    _XtTranslateOp	 opType;
{
    EventSeqPtr		 oldEventSeq;
    EventSeqPtr		 newEventSeq;
    EventSeqPtr		 lastOldEventSeq;
    EventSeqPtr		 top = NULL;
    EventSeqPtr		 last = NULL;
    EventSeqPtr		 create;

    if ( !old && !new )
	return NULL;
    if ( !old || !old->eventSeq )
	return DupAllEventSeq( new->eventSeq );
    if ( !new || !new->eventSeq )
	return DupAllEventSeq( old->eventSeq );

    oldEventSeq = old->eventSeq;
    newEventSeq = new->eventSeq;

    for ( ; newEventSeq != NULL; newEventSeq = newEventSeq->next_level ) {
	for ( oldEventSeq = old->eventSeq, lastOldEventSeq = NULL;
		oldEventSeq != NULL; lastOldEventSeq = oldEventSeq,
			oldEventSeq = oldEventSeq->next_level ) {
	    if ( CheckEventSeq( oldEventSeq, newEventSeq ) ) {
		if ( opType == XtTableOverride ) {
		    create = DupLinkedEventSeq( newEventSeq );
		} else {
		    create = DupLinkedEventSeq( oldEventSeq );
		}
		if ( !top ) {
		    top = last = create;
		} else {
		    last->next_level = create;
		    last = last->next_level;
		}
		if ( !lastOldEventSeq ) {
		    old->eventSeq = oldEventSeq->next_level;
		    oldEventSeq->next_level = NULL;
		    freeEventSeq( oldEventSeq );
		} else {
		    lastOldEventSeq->next_level = oldEventSeq->next_level;
		    oldEventSeq->next_level = NULL;
		    freeEventSeq( oldEventSeq );
		}
		break;
	    }
	}
	if ( !top ) {
	    top = last = DupLinkedEventSeq( newEventSeq );
	} else {
	    last->next_level = DupLinkedEventSeq( newEventSeq );
	    last = last->next_level;
	}
    }
    for ( oldEventSeq = old->eventSeq;
		oldEventSeq != NULL; oldEventSeq = oldEventSeq->next_level )
	if ( !top ) {
	    top = last = DupLinkedEventSeq( oldEventSeq );
	} else {
	    last->next_level = DupLinkedEventSeq( oldEventSeq );
	    last = last->next_level;
	}
    return top;
}

Private HttTranslations
MergeTranslationTable( tc, old, new, opType )
    HttTCid		tc;
    HttTranslations	old;
    HttTranslations	new;
    _XtTranslateOp	opType;
{
    HttTranslations	xlations;

    if ( !old  &&  !new )
	return NULL;
    if ( !old )
	return DupTranslations( new );
    if ( !new )
	return DupTranslations( old );
    if ( opType == XtTableReplace ) {
	return DupTranslations( new );
    }
    xlations = XtNew( TranslationData );
    xlations->operation = old->operation;
    xlations->parseTree = XtNew( TMParseStateTreeRec );
    tc->isStackQuarks = False;
    xlations->parseTree->eventSeq = MergeTranslationProc( tc,
				old->parseTree, new->parseTree, opType );
    return xlations;
}

Private HttTranslations
MergeTranslations(tc,old,new)
    HttTCid	 	tc;
    HttTranslations	old;
    HttTranslations	new;
{
    if ( !new )
	return MergeTranslationTable(tc,old,new,XtTableReplace);
    else
	return MergeTranslationTable(tc,old,new,new->operation);
}

#define OverrideTranslations(tc,old,new)	\
			MergeTranslationTable(tc,old,new,XtTableOverride)

Private HttTranslations
GetClassTranslations( tc, name, class )
    HttTCid	 tc;
    String	 name;
    String	 class;
{
    HttTranslations	default_translations;
    HttTranslations	xdefault_translations;
    HttTranslations	user_translations;
    HttTranslations	merge_translations;

    default_translations = GetDefaultTranslations( tc, name, class );
    xdefault_translations = GetXDefaultTranslations( tc, name, class );
    merge_translations = MergeTranslations( tc,
				default_translations, xdefault_translations );
    user_translations = GetUserTranslations( tc, name, class );
    merge_translations = MergeTranslations( tc,
				merge_translations, user_translations );
    freeTranslations( default_translations );
    freeTranslations( xdefault_translations );
    freeTranslations( user_translations );
    return merge_translations;
}

Private void
NewTranslation( tc, bindName, bindClass )
    HttTCid	 tc;
    String	 bindName;
    String	 bindClass;
{
    HttTranslations	base_translations;
    HttTranslations	tmp_translations;
    char		resName[64];
    char		resClass[64];

    (void )sprintf( resName, "%s.preedit", bindName );
    (void )sprintf( resClass, "%s.Preedit", bindClass );
    base_translations = GetClassTranslations( tc, resName, resClass );

    (void )sprintf( resName, "%s.preedit.edit", bindName );
    (void )sprintf( resClass, "%s.Preedit.Edit", bindClass );
    tmp_translations = GetClassTranslations( tc, resName, resClass );
    tc->parse_table[TM_ParseState_PreeditEdit] =
		OverrideTranslations( tc, base_translations, tmp_translations );
    freeTranslations( tmp_translations );

    (void )sprintf( resName, "%s.preedit.conv", bindName );
    (void )sprintf( resClass, "%s.Preedit.Conv", bindClass );
    tmp_translations = GetClassTranslations( tc, resName, resClass );
    tc->parse_table[TM_ParseState_PreeditConv] =
		OverrideTranslations( tc, base_translations, tmp_translations );
    freeTranslations( tmp_translations );

    (void )sprintf( resName, "%s.lookup.choice", bindName );
    (void )sprintf( resClass, "%s.Lookup.Choice", bindClass );
	    tc->parse_table[TM_ParseState_LookupChoice] =
			GetClassTranslations( tc, resName, resClass );

    freeTranslations( base_translations );
    return;
}

typedef struct _tc_classRec *	tc_class;
typedef struct _tc_classRec {
    tc_class		 next;
    int			 usecount;
    char		 bindName[64];
    HttTranslations      parse_table[TM_MaxParseState_Count];
    short		 numQuarks;
    short		 quarkTblSize;
    XrmQuark		*quarkTbl;
} tc_classRec;

Private tc_class	tc_class_queue_terminal = NULL;

Private void
MakeTranslation( tc, bindName, bindClass )
    HttTCid	 tc;
    String	 bindName;
    String	 bindClass;
{
    tc_class	 tcc = tc_class_queue_terminal;
    tc_class	 ltcc = tc_class_queue_terminal;

    strcpy( tc->bindName, bindName );
    if ( tcc ) {
	for ( ; tcc; ltcc = tcc, tcc = tcc->next ) {
	    if ( !strcmp( tcc->bindName, bindName ) ) {
		tcc->usecount++;
		tc->parse_table[TM_ParseState_PreeditEdit] =
				tcc->parse_table[TM_ParseState_PreeditEdit];
		tc->parse_table[TM_ParseState_PreeditConv] =
				tcc->parse_table[TM_ParseState_PreeditConv];
		tc->parse_table[TM_ParseState_LookupChoice] =
				tcc->parse_table[TM_ParseState_LookupChoice];
		tc->numQuarks = tcc->numQuarks;
		tc->quarkTblSize = tcc->quarkTblSize;
		tc->quarkTbl = tcc->quarkTbl;
		return;
	    }
	}
	ltcc->next = tcc = XtNew(tc_classRec);
    } else {
	tc_class_queue_terminal = tcc = XtNew(tc_classRec);
    }
    tcc->usecount = 1;
    strcpy( tcc->bindName, bindName );
    NewTranslation( tc, bindName, bindClass );
    tcc->parse_table[TM_ParseState_PreeditEdit] =
		tc->parse_table[TM_ParseState_PreeditEdit];
    tcc->parse_table[TM_ParseState_PreeditConv] =
		tc->parse_table[TM_ParseState_PreeditConv];
    tcc->parse_table[TM_ParseState_LookupChoice] =
		tc->parse_table[TM_ParseState_LookupChoice];
    tcc->numQuarks = tc->numQuarks;
    tcc->quarkTblSize = tc->quarkTblSize;
    tcc->quarkTbl = tc->quarkTbl;
    return;
}

Private void
DestroyTranslation( tc )
    HttTCid	tc;
{
    int		 i;
    tc_class	 tcc = tc_class_queue_terminal;
    tc_class	 ltcc = tc_class_queue_terminal;

    if ( !tcc )
	return;
    for ( ; tcc; ltcc = tcc, tcc = tcc->next ) {
	if ( !strcmp( tcc->bindName, tc->bindName ) ) {
	    if ( tcc->usecount > 0 ) {
		tcc->usecount--;
	    }
#ifdef NeedAlwaysMakeTranslations
	    if ( tcc->usecount != 0 ) {
		return;
	    }
	    for ( i = 0; i < TM_MaxParseState_Count; i++ )
		freeTranslations( tc->parse_table[i] );
	    XtFree( tc->quarkTbl );
	    if ( tcc == tc_class_queue_terminal ) {
		tc_class_queue_terminal = tcc->next;
	    } else {
		ltcc->next = tcc->next;
	    }
	    XtFree( tcc );
#endif /* NeedAlwaysMakeTranslations */
	    return;
	}
    }
    return;
}

Private void
_TranslationInitialized()
{
    if ( initialized )
	return;

    initialized = True;

    Compile_XtModifierTable( modifiers, XtNumber(modifiers) );
    CompileNameValueTable( buttonNames );
    CompileNameValueTable( motionDetails );
    CompileNameValueTable( notifyModes );
    CompileNameValueTable( mappingNotify );
    Compile_XtEventTable( events, XtNumber(events) );
}

Private int
_XtGetQuarkIndex( tc, quark )
    HttTCid		tc;
    XrmQuark		quark;
{
#define TM_QUARK_TBL_ALLOC	16
#define TM_QUARK_TBL_REALLOC	16
    register int	i = tc->numQuarks;

    for ( i = 0; i < tc->numQuarks; i++ )
	if ( tc->quarkTbl[i] == quark )
	    break;
    if ( i == tc->numQuarks ) {
	if ( tc->numQuarks == tc->quarkTblSize ) {
	    int		newSize;

	    if ( tc->quarkTblSize == 0 )
		tc->quarkTblSize += TM_QUARK_TBL_ALLOC;
	    else
		tc->quarkTblSize += TM_QUARK_TBL_REALLOC;
	    newSize = ( tc->quarkTblSize * sizeof(XrmQuark) );

	    if ( tc->isStackQuarks ) {
		XrmQuark	*oldQuarkTbl = tc->quarkTbl;
		tc->quarkTbl = (XrmQuark *)XtMalloc( newSize );
		XtBCopy( oldQuarkTbl, tc->quarkTbl, newSize );
		tc->isStackQuarks = False;
	    } else {
		tc->quarkTbl = (XrmQuark *)
			XtRealloc( (char *)tc->quarkTbl,
				( tc->quarkTblSize *
				  sizeof(XrmQuark) ) );
	    }
	}
	tc->quarkTbl[tc->numQuarks++] = quark;
    }
    return i;
}

Private void
freeCurrentEventSeq( curEventSeq )
    TMEventSeqPtr	curEventSeq;
{
    TMEventSeqPtr	nextEventSeq;

    for ( ; curEventSeq; curEventSeq = nextEventSeq ) {
	nextEventSeq = curEventSeq->next;
	XtFree( curEventSeq->tmEvent.xev );
	XtFree( curEventSeq );
    }
    return;
}

Private CompileAction *
GetCompileAction( tc, actionQuark )
    HttTCid	tc;
    XrmQuark	actionQuark;
{
    int		i;
    ActionList	actionList = tc->action_table;

    for ( ; actionList; actionList = actionList->next ) {
	CompiledActionTable	cAction = actionList->table;
	for ( i = 0; i < (int )actionList->count; i++, cAction++ ) {
	    if ( cAction->signature == actionQuark )
		return cAction;
	}
    }
    return (CompileAction *)NULL;
}

Private void
CallMatchActions( tc, last_es, action )
    HttTCid		tc;
    TMEventSeqPtr	last_es;
    ActionPtr		action;
{
    int			 i;
    CompileAction	*cAction;

    for ( ; action; action = action->next ) {
	cAction = GetCompileAction(
			tc, TMGetActionQuark( tc->quarkTbl, action->idx ) );
	if ( cAction ) {
	    ( cAction->proc )( tc->sc, last_es->tmEvent.xev,
				action->params, action->num_params );
	}
    }
    return;
}

Private void
CallUnmatchAction( tc, tmSeq )
    HttTCid		tc;
    TMEventSeqPtr	tmSeq;
{
    if ( !tc->unmatchProc )
	return;
    (tc->unmatchProc)( tc->sc, tmSeq->tmEvent.xev, tc->unmatchParams, tc->unmatchNum );
    return;
}

Private Boolean
CheckUnmatchCase( tc, last_es, first_match, middle_match )
    HttTCid	tc;
    TMEventSeqPtr	last_es;
    Boolean		first_match;
    Boolean		middle_match;
{
    TMEventSeqPtr	next = tc->curEventSeq->next;

    if ( !first_match || !middle_match ) {
	if ( next ) {
	    CallUnmatchAction( tc, next );
	    free( tc->curEventSeq );
	    tc->curEventSeq = next;
	    (void)TMMatchEvent( tc, tc->translation_mode );
	} else {
	    tc->curEventSeq = (TMEventSeqPtr )NULL;
	    CallUnmatchAction( tc, last_es );
	    free( last_es );
	}
    }
    return False;
}

#define MatchIncommingEvent(tmStep,curSeq) \
	    (tmStep->event.eventType == curSeq->tmEvent.event.eventType && \
	      (tmStep->event.matchEvent != NULL ) && \
	      (*tmStep->event.matchEvent)(tmStep, curSeq))

Private Boolean
TMKeyMatch( tc, tmseq, last_es, unmatchNotify )
    HttTCid		tc;
    EventSeqPtr		tmseq;
    TMEventSeqPtr	last_es;
    Boolean		unmatchNotify;
{
    Boolean		first_match = False;
    Boolean		middle_match = False;

    for ( ; tmseq; tmseq = tmseq->next_level ) {
	EventSeqPtr	tmstep = tmseq;
	EventSeqPtr	tmlast = tmseq;
	TMEventSeqPtr	curseq;
	for ( curseq = tc->curEventSeq; curseq;
	    curseq = curseq->next, tmlast = tmstep, tmstep = tmstep->next ) {
	    if ( MatchIncommingEvent( tmstep, curseq ) ) {
		first_match = True;
	    } else break;
	}
	if ( !curseq && !tmstep ) {
	    /*
	     * Match key event sequence
	     */
	    CallMatchActions( tc, last_es, tmlast->actions );
	    freeCurrentEventSeq( tc->curEventSeq );
	    tc->curEventSeq = (TMEventSeqPtr )NULL;
	    return True;
	}
	if ( !curseq ) {
	    middle_match = True;
	}
    }
    /*
     * Unmatch key event sequence
     */
    if ( !unmatchNotify ) {
	if ( !middle_match ) {
	    freeCurrentEventSeq( tc->curEventSeq );
	    tc->curEventSeq = (TMEventSeqPtr )NULL;
	}
    } else
	return CheckUnmatchCase( tc, last_es, first_match, middle_match );
    return False;
}

Private Boolean
TMMatchEvent( tc, mode )
    HttTCid	 tc;
    int		 mode;
{
    TMEventSeqPtr	es  = tc->curEventSeq;
    TMEventSeqPtr	les = tc->curEventSeq;
    EventSeqPtr		tmseq;

    if ( mode < 0 || mode >= TM_MaxParseState_Count )
	return True;
    if ( !tc || !tc->parse_table[mode] || !tc->parse_table[mode]->parseTree ||
	 !tc->parse_table[mode]->parseTree->eventSeq )
	return True;
    if ( !es )
	return True;
    tc->translation_mode = mode;
    tmseq = tc->parse_table[mode]->parseTree->eventSeq;
    for ( ; es; les = es, es = es->next );
    return TMKeyMatch( tc, tmseq, les, True );
}

Private void
TMAddEventSeq( tc, e )
    HttTCid	 tc;
    XEvent	*e;
{
    TMEventSeqPtr	es  = tc->curEventSeq;
    TMEventSeqPtr	les = tc->curEventSeq;

    if ( es ) {
	for ( ; es; les = es, es = es->next );
	es = les->next = (TMEventSeqPtr)XtMalloc( sizeof(EventSeqRec) );
    } else {
	es = tc->curEventSeq = (TMEventSeqPtr)XtMalloc( sizeof(EventSeqRec) );
    }
    XEventToTMEvent( tc, e, &es->tmEvent );
    return;
}

Private void
TMClearEventSeq( tc )
    HttTCid	tc;
{
    freeCurrentEventSeq( tc->curEventSeq );
    tc->curEventSeq = (TMEventSeqPtr )NULL;
    return;
}

Private Boolean
TMEventToString( tc, str, strLen, keysym )
    HttTCid	 tc;
    String	 str;
    int		 strLen;
    KeySym	*keysym;
{
    if ( tc->curEventSeq ) {
	tc->sc->If->m->iml_lookupString( tc->curEventSeq->tmEvent.xev,
				str, strLen, keysym );
	return True;
    } else return False;
}

/*
 * Action Proc
 */
Private CompiledActionTable
CompileActionTable(actions, count, stat, perm)
    register HttActionRec *actions;
    register Cardinal count;	/* may be 0 */
    Boolean stat;	/* if False, copy before compiling in place */
    Boolean perm;	/* if False, use XrmStringToQuark */
{
    register CompiledActionTable cActions;
    register int i;
    CompileAction hold;
    CompiledActionTable cTableHold;
    XrmQuark (*func)();

    if (!count)
	return (CompiledActionTable) NULL;
    func = (perm ? XrmPermStringToQuark : XrmStringToQuark);

    if (! stat) {
	cTableHold = cActions = (CompiledActionTable)
	    XtMalloc(count * sizeof(CompileAction));

	for (i=count; --i >= 0; cActions++, actions++) {
	    cActions->proc = actions->proc;
	    cActions->signature = (*func)(actions->string);
	}
    } else {
	cTableHold = (CompiledActionTable) actions;

	for (i=count; --i >= 0; actions++)
	    ((CompiledActionTable) actions)->signature = 
		(*func)(actions->string);
    }
    cActions = cTableHold;

    /* Insertion sort.  Whatever sort is used, it must be stable. */
    for (i=1; i <= count - 1; i++) {
	register int j;
	hold = cActions[i];
	j = i;
	while (j && cActions[j-1].signature > hold.signature) {
	    cActions[j] = cActions[j-1];
	    j--;
	}
	cActions[j] = hold;
    }

    return cActions;
}

Private void
AddActions( tc, actions, num_actions )
    HttTCid		 tc;
    HttActionList	 actions;
    Cardinal		 num_actions;
{
    register ActionList rec;

    if ( !tc )
	return;

    rec = XtNew(ActionListRec);
    rec->next = tc->action_table;
    tc->action_table = rec;
    rec->table = CompileActionTable(actions, num_actions, False, False);
    rec->count = num_actions;
    return;
}

Private void
AddUnmatchAction( tc, proc, params, num_params )
    HttTCid	  tc;
    void	(*proc)(/* tc, event */);
    String	 *params;
    int		  num_params;
{
    int		i;

    if ( tc->unmatchParams ) {
	/* Already regist. */
	for ( i = 0; i < tc->unmatchNum; i++ )
	    XtFree( tc->unmatchParams[i] );
	XtFree( tc->unmatchParams );
	tc->unmatchParams = (String *)NULL;
    }
    tc->unmatchProc = proc;
    if ( num_params != 0 ) {
	String *list = (String *)XtMalloc( num_params * sizeof(String) );
	for ( i = 0; i < num_params; i++ ) {
	    int	    size = strlen( params[i] );
	    list[i] = (String )XtRealloc( (char *)params[i], size + 1 );
	}
	tc->unmatchParams = list;
    }
    tc->unmatchNum = num_params;
    return;
}

/*
 * TranslationContext Management
 */

Private void
makeTMMethodRec( tc )
    HttTCid	tc;
{
    tc->tm = XtNew( TMMethodRec );
    tc->tm->tm_make_translations = MakeTranslation;
    tc->tm->tm_match_event = TMMatchEvent;
    tc->tm->tm_add_eventSeq = TMAddEventSeq;
    tc->tm->tm_clear_eventSeq = TMClearEventSeq;
    tc->tm->tm_add_action = AddActions;
    tc->tm->tm_unmatch_action = AddUnmatchAction;
    tc->tm->tm_event_to_String = TMEventToString;
    return;
}

Private	HttTCRec	*tc_queue_terminal = NULL;

#if NeedFunctionPrototypes
Public HttTCid
Htt_CreateTC( void *sc )
#else
Public HttTCid
Htt_CreateTC( sc )
    void	*sc;
#endif
{
    HttTCRec	*tc = tc_queue_terminal;
    HttTCRec	*ltc = tc_queue_terminal;

    if ( tc ) {
	for ( ; tc; ltc = tc, tc = tc->next );
	tc = ltc->next = (HttTCRec *)XtMalloc( sizeof(HttTCRec) );
    } else {
	_TranslationInitialized();
	tc = tc_queue_terminal = (HttTCRec *)XtMalloc( sizeof(HttTCRec) );
    }
    bzero( tc, sizeof(HttTCRec) );
    tc->sc = sc;
    makeTMMethodRec( tc );
    return (HttTCid)tc;
}

#if NeedFunctionPrototypes
Public void
Htt_DestroyTC( HttTCid closed_tc_id )
#else
Public void
Htt_DestroyTC( closed_tc_id )
    HttTCid	closed_tc_id;
#endif
{
    int		 i;
    HttTCRec	*tc = tc_queue_terminal;
    HttTCRec	*ltc = tc_queue_terminal;

    if ( tc ) {
	XtFree( tc->tm );
	DestroyTranslation( tc );
	freeActionList( tc->action_table );
	freeCurrentEventSeq( tc->curEventSeq );
	for ( i = 0; i < tc->unmatchNum; i++ )
	    XtFree( tc->unmatchParams[i] );
	XtFree( tc->unmatchParams );
	for ( ; tc; ltc = tc, tc = tc->next ) {
	    if ( (HttTCid)tc == closed_tc_id ) {
		ltc->next = tc->next;
		XtFree( tc );
		return;
	    }
	}
    }
    return;
}
